package org.netkernel.servlet.http;

import org.netkernel.layer0.bnf.BNFGrammarFactory;
import org.netkernel.layer0.bnf.IGrammarGroup;
import org.netkernel.layer0.meta.IIdentifierGrammar;
import org.netkernel.layer0.meta.ISpaceElements;
import org.netkernel.layer0.meta.ISpaceMeta;
import org.netkernel.layer0.meta.impl.EndpointMetaImpl;
import org.netkernel.layer0.meta.impl.SimpleSpaceElementsImpl;
import org.netkernel.layer0.nkf.*;
import org.netkernel.layer0.representation.ByteArrayRepresentation;
import org.netkernel.layer0.representation.IBinaryStreamRepresentation;
import org.netkernel.layer0.representation.IReadableBinaryStreamRepresentation;
import org.netkernel.layer0.urii.SimpleIdentifierImpl;
import org.netkernel.layer0.util.RegexRewriterRule;
import org.netkernel.layer0.util.Utils;
import org.netkernel.layer0.util.XMLReadable;
import org.netkernel.module.standard.endpoint.ConfiguredOverlayImpl.IConfig;
import org.netkernel.module.standard.endpoint.IStandardEndpoint;
import org.netkernel.request.IRequest;
import org.netkernel.request.IRequestScopeLevel;
import org.netkernel.request.impl.RequestScopeLevelImpl;
import org.netkernel.servlet.http.rep.IHttpRedirect;
import org.netkernel.urii.IEndpoint;
import org.netkernel.urii.IMetaRepresentation;
import org.netkernel.urii.INetKernelException;
import org.netkernel.urii.ISpace;
import org.netkernel.urii.impl.NetKernelError;
import org.netkernel.urii.impl.NetKernelException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import javax.servlet.http.Cookie;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;


public class HTTPBridgeOverlayConfig implements IConfig
  {
  private final RegexRewriterRule[] mRules;
  private boolean mTranslateHTTPMethods = false;
  private boolean mNegotiateContent = false;
  private String mContentSignalType = "header";
  private final ISpace mDelegateSpace;
  private final String mExceptionHandler;
  private final IEndpoint mEndpoint;
  private final String mElementId;
  private final EndpointMetaImpl mMeta;

  public static final String HTTP_CACHE_EXPIRES_DELTA_SECONDS = "HTTP_CACHE_EXPIRES_DELTA_SECONDS";
//  public static final String HTTPBRIDGE_TRANSPARENT_PASS_THROUGH = "HTTPBRIDGE_TRANSPARENT_PASS_THROUGH";


  public HTTPBridgeOverlayConfig(ISpace aDelegateSpace, ISpaceMeta aDelegateMeta, INKFRequestContext aContext, IEndpoint aEndpoint) throws Exception
    {
    mDelegateSpace = aDelegateSpace;
    mExceptionHandler = (String) aContext.getParamValue("exceptionHandler");
    mEndpoint = aEndpoint;
    Document config = null;
    if (aContext.getParamValue("config") != null)
      {
      config = aContext.source("param:config", Document.class);
      }
    IGrammarGroup grammarRoot = BNFGrammarFactory.createChoiceGroup(null);
    if (config != null)
      {
      XMLReadable c = new XMLReadable(config);

      // Process re-write rules
      List<RegexRewriterRule> rules = new ArrayList<RegexRewriterRule>();
      for (Iterator<Node> i = c.getNodes("/config/rewrite").iterator(); i.hasNext();)
        {
        XMLReadable rewrite = new XMLReadable(i.next());
        String match = rewrite.getText("match");
        String to = rewrite.getText("to");
        RegexRewriterRule rule = new RegexRewriterRule(match, to);
        rules.add(rule);
        BNFGrammarFactory.createRegex(grammarRoot, match);
        }
      mRules = new RegexRewriterRule[rules.size()];
      rules.toArray(mRules);

      // Determine if we should translate HTTP Methods to ROC verbs
      c = new XMLReadable(config);
      for (Iterator<Node> i = c.getNodes("/config").iterator(); i.hasNext();)
        {
        String translate = (new XMLReadable(i.next())).getText("translateHTTPMethods");
        mTranslateHTTPMethods = "true".equals(translate);
        }

      // Determine if we should do content negotiation
      c = new XMLReadable(config);
      for (Iterator<Node> i = c.getNodes("/config").iterator(); i.hasNext();)
        {
        mContentSignalType = (new XMLReadable(i.next())).getText("negotiateContent");
        if ("header".equals(mContentSignalType) || "argument".equals(mContentSignalType))
          {
          mNegotiateContent = true;
          }
        }
      }
    else
      {
      mRules = new RegexRewriterRule[1];
      mRules[0] = new RegexRewriterRule("(htt(p|ps)://.*)", "$1");
      mTranslateHTTPMethods = false;
      }


    //setup meta
    mElementId = Utils.generateUniqueEndpointIdentifier(mDelegateSpace, aEndpoint, "HTTPBridgeOverlayConfig");
    IIdentifierGrammar grammar = BNFGrammarFactory.createGrammar(grammarRoot, aContext.getKernelContext().getKernel().getLogger());
    mMeta = new EndpointMetaImpl("HTTPBridge", null, null, null, IRequest.VERB_SOURCE, grammar, EndpointMetaImpl.NO_ARGS, EndpointMetaImpl.NO_RESPONSES, EndpointMetaImpl.NO_EXCEPTIONS);
    mMeta.addAdditionalField("private", true);

    System.out.println("Translating HTTP Methods [" + mTranslateHTTPMethods + "]");
    System.out.println("Negotiating Content [" + mNegotiateContent + "]");
    }

  public void onRequest(String aID, INKFRequestContext context) throws Exception
    {
    System.out.println("onRequest() at HTTPBridge");

    Object[] a = (Object[]) context.getThisRequest().getPrimary();
    HttpServletRequest httpRequest = (HttpServletRequest) a[0];
    HttpServletResponse httpResponse = (HttpServletResponse) a[1];


    //Perform any inner request rewriting
    String innerID = context.getThisRequest().getIdentifier();
    innerID = Utils.decode(innerID);
    if (mRules.length > 0)
      {
      for (RegexRewriterRule rule : mRules)
        {
        if (rule.matches(innerID))
          {
          innerID = rule.map(innerID);
          }
        }
      }

    // Now prepare the inner request for injected into the prepared context
    // Note that we either issue a simple request or iterate for content negotiation
    INKFResponseReadOnly innerResponse = null;
    INKFRequest request;
    HTTPRequestSpace httpRequestSpace = null;
    IRequestScopeLevel injectionScope = null;
    try
      {

      // We only negotiate content if requested and the HTTP method is either GET or HEAD
      String httpMethod = httpRequest.getMethod();
      if (! (mNegotiateContent && ("GET".equals(httpMethod) || "HEAD".equals(httpMethod)) ))
        {
        System.out.println("No content negotiation");
        // Create the injected space context and then issue the request
        httpRequestSpace = new HTTPRequestSpace(httpRequest);
        injectionScope = RequestScopeLevelImpl.createDurableScopeLevel(httpRequestSpace, context.getKernelContext().getThisKernelRequest().getRequestScope());
        injectionScope = RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, injectionScope);
        httpRequestSpace.onCommissionSpace(context.getKernelContext().getKernel());
        request = context.createRequest(innerID);
        request.setRequestScope(injectionScope);
        request.setRepresentationClass(IBinaryStreamRepresentation.class);
        setVerb(context, request, httpRequest);
        innerResponse = context.issueRequestForResponse(request);
        }
      else
        {
        System.out.println("Yes, we are doing content negotiation");
        // With content negotiation, we examine the "Accept" header and create a request for a highest priority MIME type.
        // If that request fails with an exception, then we pick the next highest priority MIME type.

        String acceptHeader = httpRequest.getHeader("Accept");
        System.out.println("The Accept header is [" + acceptHeader + "]");
        MimeTypePriority[] mimeTypes = parseAcceptHeader(acceptHeader);

        boolean requestFailed = true;
        boolean mimeTypeAvailable = true;

        for (int i = 0; i < mimeTypes.length && requestFailed && mimeTypeAvailable; i++)
          {
          MimeTypePriority mimeType = mimeTypes[i];
          System.out.println(mimeType);

          if (mimeType.isEndValue())
            {
            System.out.println("We have an endvalue");
            // Final attempt
            if (0 < mimeType.getPriority())
              {
              // In this case we don't care about the mimetype, just get whatever resource comes back
              // Create the injected space context and then issue the request
              httpRequestSpace = new HTTPRequestSpace(httpRequest);
              injectionScope = RequestScopeLevelImpl.createDurableScopeLevel(httpRequestSpace, context.getKernelContext().getThisKernelRequest().getRequestScope());
              injectionScope = RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, injectionScope);
              httpRequestSpace.onCommissionSpace(context.getKernelContext().getKernel());
              request = context.createRequest(innerID);
              request.setRequestScope(injectionScope);
              request.setRepresentationClass(IBinaryStreamRepresentation.class);
              setVerb(context, request, httpRequest);
              innerResponse = context.issueRequestForResponse(request);
              mimeTypeAvailable = false;
              }
            else
              {
              // In this case we have no other options and we must fail
              request = context.createRequest("active:failureResult");
              mimeTypeAvailable = false;
              }

            }
          else
            {
            System.out.println("We don't have an endvalue");
            // Create a request that includes information about the Mime Type we are interested in

            // Create the injected space context and then issue the request
            httpRequestSpace = new HTTPRequestSpace(httpRequest);
            injectionScope = RequestScopeLevelImpl.createDurableScopeLevel(httpRequestSpace, context.getKernelContext().getThisKernelRequest().getRequestScope());
            injectionScope = RequestScopeLevelImpl.createNonDurableScopeLevel(mDelegateSpace, injectionScope);
            System.out.println("Before space commission");
            httpRequestSpace.onCommissionSpace(context.getKernelContext().getKernel());
            System.out.println("After space commission");
            request = context.createRequest(innerID);
            request.setRequestScope(injectionScope);
            request.setRepresentationClass(IBinaryStreamRepresentation.class);
            setVerb(context, request, httpRequest);
            System.out.println("Content signal [" + mContentSignalType +"]");
            if ("header".equals(mContentSignalType))
              {
              // Use a sticky header
              request.setHeader("mimetype", mimeType.getMimeType(), true);
              System.out.println("Set sticky header mimetype to [" + mimeType.getMimeType());
              }
            else
              {
              // Use a named argument
              request.addArgument("mimetype", mimeType.getMimeType());
              System.out.println("Set argument mimetype to [" + mimeType.getMimeType() + "]");
              }

            }
          try
            {
            //Commission the prepared context
            // We are now ready to issue the inner sub-request.
            innerResponse = context.issueRequestForResponse(request);
            requestFailed = false;
            }
          catch (NetKernelException e)
            {
            System.out.println("Exception ID: [" + e.getDeepestId() +"]");
            if ("Request Resolution Failure".equals(e.getDeepestId()) | "Unsupported Mime type".equals(e.getDeepestId()))
              {
              System.out.println("Proposed MIME type rejected");
              }
            else
              {
              System.out.println("Re-throwing Exception");
              throw e;
              }
            }
          }
        System.out.println("Request failed [" + requestFailed + "]");
        if (requestFailed)
          {
          // We did not get back the MIME type that we need
          // Send back an HttpServletResponse to indicate this
          httpResponse.setStatus(406);
          return;
          }
        }

      // TODO: This is the only place that needs handling is called. Can we just do this on the stack?
      // Why is this here? what condition is it protecting?
//      if (!httpRepresentation.needsHandling())
//        {  //Response has already been written by the HTTP request handler timeout code.
//        INKFResponse response = context.createResponseFrom(null);
//        response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
//        return;
//        }
      IBinaryStreamRepresentation innerRepresentation = (IBinaryStreamRepresentation) innerResponse.getRepresentation();
      httpRequestSpace.applyResponseMetadata(innerResponse.getHeaders(), context);

      if (innerResponse.hasHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER))
        {  //Handle Redirects
        IHttpRedirect redirect = (IHttpRedirect) innerResponse.getHeader(IHttpRedirect.HTTP_REDIRECT_METADATA_IDENTIFIER);
        httpResponse.setStatus(redirect.getType());
        // TODO: Change to constant
        httpResponse.setHeader("Location", redirect.getURL());
        if (redirect.getType() == IHttpRedirect.HTTP_REDIRECT_300_MULTIPLE_CHOICES)
          {  //Write choices...
          httpResponse.setContentType("text/plain");
          OutputStreamWriter w = new OutputStreamWriter(httpResponse.getOutputStream());
          for (String s : redirect.getChoices())
            {
            w.write(s);
            w.write("\n");
            }
          w.close();
          }
        else if (innerRepresentation != null)
          {  //Comply with HTTP1.1 spec and send short note as body if it is present.
          writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRequest);
          }
        }
      else
        {
        ////////////////////////////////////////
        //Process HTTP Response
        ////////////////////////////////////////
        boolean writestream = writeHeaders(innerResponse, httpResponse, httpRequestSpace);
        if (innerRepresentation != null && writestream)
          {  //Etag Processing
          String method = httpRequest.getMethod();
          // TODO: Change to constants
          if ((method.equals("GET") || method.equals("HEAD")) && !innerResponse.isExpired())
            {
            //TODO: Change to constant
            String etag = httpRequest.getHeader("If-None-Match");  //Doesn't matter if null
            String tagString = "\"" + Integer.toHexString(innerRepresentation.hashCode()) + "\"";
            //Write the Etag
            //TODO: Change to constant
            httpResponse.setHeader("ETag", tagString);
            if (tagString.equals(etag))
              {  //Not modified
              writestream = false;
              httpResponse.setStatus(HttpServletResponse.SC_NOT_MODIFIED);
              }
            }
          //Write the representation.
          // TODO: Change to constant
          if (writestream && !method.equals("HEAD"))
            {
            //TODO: Change to constant
            httpResponse.setDateHeader("Last-Modified", System.currentTimeMillis());
            writeStream(innerRepresentation, httpResponse, innerResponse.getMimeType(), httpRequest);
            }
          }
        else if (writestream)
          {
          throw context.createFormattedException("EX_NULL_RESPONSE", "MSG_NULL_RESPONSE", null, null);
          }
        }
      }
//			catch(org.mortbay.jetty.EofException e)
//			{	//Requestor has gone away so no point handling Exception.
//				context.logRaw(INKFRequestContext.LEVEL_FINE, e.getMessage());
//				context.logRaw(INKFRequestContext.LEVEL_FINER, e.toString());
//			}
    catch (NKFException e)
      {
      handleException(e, context, injectionScope, httpResponse, httpRequest, httpRequestSpace);
      }
    catch (NetKernelError e)
      {
      handleException(e, context, injectionScope, httpResponse, httpRequest, httpRequestSpace);
      }
    catch (Exception e)
      {
      INetKernelException ex = context.createFormattedException("EX_IN_HTTP_BRIDGE", null, null, e);
      handleException(ex, context, injectionScope, httpResponse, httpRequest, httpRequestSpace);
      }
    INKFResponse response = context.createResponseFrom(null);
    response.setExpiry(INKFResponse.EXPIRY_ALWAYS);
    }

  private boolean writeHeaders(INKFResponseReadOnly innerResponse, HttpServletResponse httpResponse, HTTPRequestSpace httpRequestSpace) throws IOException, NKFException
    {
    boolean writestream = true;

    //TODO: Constant and Configuration based
    httpResponse.setHeader("Server", "NetKernel 4.0.2");

    //Process Cookies
    Map cookies = httpRequestSpace.getResponseCookies();
    Iterator i = cookies.keySet().iterator();
    while (i.hasNext())
      {
      Cookie c = (Cookie) cookies.get(i.next());
      httpResponse.addCookie(c);
      }

    //Process Headers
    Map headers = httpRequestSpace.getResponseHeaders();
    i = headers.keySet().iterator();
    while (i.hasNext())
      {
      String name = (String) i.next();
      Object value = headers.get(name);
      if (value instanceof Long)
        {
        httpResponse.setDateHeader(name, (Long) value);
        }
      else if (value instanceof Integer)
        {
        httpResponse.setIntHeader(name, (Integer) value);
        }
      else
        {
        httpResponse.setHeader(name, value.toString());
        }
      }

    if (innerResponse.hasHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS))
      {
      String vs = innerResponse.getHeader(HTTP_CACHE_EXPIRES_DELTA_SECONDS).toString();
      try
        {
        // TODO: constant
        httpResponse.setDateHeader("Expires", System.currentTimeMillis() + Long.parseLong(vs) * 1000L);
        }
      catch (Exception e)
        {
        e.printStackTrace();  //Wrong number format.
        }
      //TODO: Constant
      httpResponse.setHeader("Cache-Control", "public, max-age=" + vs);
      }

    //Process Response Code
    httpResponse.setStatus(httpRequestSpace.getResponseCode());

    //Process Redirect
    if (httpRequestSpace.shouldRedirect())
      {
      httpResponse.sendRedirect(httpRequestSpace.getResponseRedirect());
      writestream = false;
      }
    return writestream;
    }

  private void writeStream(IBinaryStreamRepresentation innerRepresentation, HttpServletResponse httpResponse, String mimeType, HttpServletRequest httpRequest) throws Exception
    {  //String mimeType=innerResponse.getMimeType();
    String encoding = innerRepresentation.getEncoding();
    if (encoding != null)
      {
      int gp = encoding.indexOf("gzip");
      if (gp >= 0)
        {
        httpResponse.setHeader("Content-Encoding", "gzip");
        if (gp > 0)
          {
          encoding = encoding.substring(0, gp - 1);    //Strip +gzip tag
          }
        else
          {
          encoding = null;                        //Prevent writing gzip encoding on mimetype header.
          }
        }
      }
    if (mimeType != null)
      {  //Special case to make XHTML work with crappy Internet Explorer
      if (mimeType.equals("application/xhtml+xml"))
        {
        // TODO: Constant
        if (httpRequest.getHeader("User-Agent").contains("MSIE"))
          {
          mimeType = "text/html";
          }
        }
      if ((encoding != null) && mimeType.startsWith("text/"))    //Encoding when text but we need a complete mimetype - encoding database see: http://www.w3.org/Protocols/rfc2616/rfc2616-sec3.html#sec3.7
        {
        mimeType += ";charset=" + encoding;
        }
      httpResponse.setContentType(mimeType);
      }
    //Content-Length
    if (innerRepresentation instanceof IReadableBinaryStreamRepresentation)
      {
      IReadableBinaryStreamRepresentation rbs = (IReadableBinaryStreamRepresentation) innerRepresentation;
      httpResponse.setContentLength(rbs.getContentLength());
      }

    OutputStream s = httpResponse.getOutputStream();
    innerRepresentation.write(s);
    //s.close();
    }

  private void handleException(Throwable e, INKFRequestContext context, IRequestScopeLevel injectionContext, HttpServletResponse httpResponse, HttpServletRequest httpRequest, HTTPRequestSpace httpRequestSpace) throws Exception
    {
    try
      {
      httpResponse.reset();
      }
    catch (IllegalStateException ies)
      {  /*Can't reset the response - already committed*/
      context.logRaw(INKFRequestContext.LEVEL_FINE, "Exception but HTTPResponse is complete cannot be reset. " + e.getMessage());
      return;
      }
    try
      {
      //explicit serialisation so that handler considers exception as data
      INKFRequest req = context.createRequest("literal:exception");
      req.setVerb(INKFRequestReadOnly.VERB_TRANSREPRESENT);
      req.addPrimaryArgument(e);
      req.setRepresentationClass(IBinaryStreamRepresentation.class);
      req.setRequestScope(injectionContext);
      Object serializedException = context.issueRequest(req);

      if (mExceptionHandler != null)
        {
        req = context.createRequest(mExceptionHandler);
        req.setRequestScope(injectionContext);
        req.addArgumentByValue("operand", serializedException);
        req.setRepresentationClass(IBinaryStreamRepresentation.class);
        INKFResponseReadOnly<IBinaryStreamRepresentation> response = context.issueRequestForResponse(req);
        httpRequestSpace.applyResponseMetadata(response.getHeaders(), context);
        boolean writestream = writeHeaders(response, httpResponse, httpRequestSpace);
        if (writestream)
          {
          writeStream(response.getRepresentation(), httpResponse, response.getMimeType(), httpRequest);
          }
        }
      else
        {
        httpResponse.setContentType("text/xml");
        OutputStream s = httpResponse.getOutputStream();
        IBinaryStreamRepresentation result = (IBinaryStreamRepresentation) serializedException;
        result.write(s);
        s.close();
        }
      }
    catch (Exception e2)
      {
      context.logRaw(INKFRequestContext.LEVEL_WARNING, e2.toString());
      if (e instanceof INetKernelException)
        {
        throw (INetKernelException) e;
        }
      else if (e instanceof NetKernelError)
        {
        throw (NetKernelError) e;
        }
      else
        {
        throw e2;
        }
      }
    }

  public IMetaRepresentation getElementMeta(String aElementId)
    {
    return mMeta;
    }

  public ISpaceElements getElements(INKFRequestContext aContext) throws Exception
    {
    ISpaceElements result = new SimpleSpaceElementsImpl(new SimpleIdentifierImpl(mElementId), mEndpoint);
    return result;
    }

  public String getResolvedElement(INKFRequestReadOnly aRequest) throws Exception
    {
    String result = null;
    String id = aRequest.getIdentifier();
    for (RegexRewriterRule rule : mRules)
      {
      if (rule.matches(id))
        {
        result = mElementId;
        break;
        }
      }
    return result;
    }

  public ISpace[] getSpaces(INKFRequestContext aContext) throws Exception
    {
    return IStandardEndpoint.NO_SPACES;
    }

  private IReadableBinaryStreamRepresentation getBodyContent(HttpServletRequest request) throws IOException
    {
    IReadableBinaryStreamRepresentation content = null;

    String mime = request.getContentType();

    if (mime == null || (mime != null && !mime.startsWith("application/x-www-form-urlencoded") && !mime.startsWith("multipart/form-data")))
      {
      if (request.getContentLength() > 0)
        {
        ByteArrayOutputStream baos = new ByteArrayOutputStream(request.getContentLength());
        Utils.pipe(request.getInputStream(), baos);
        content = new ByteArrayRepresentation(baos, request.getCharacterEncoding());
        }
      }

    return content;
    }

  private void setVerb(INKFRequestContext context, INKFRequest request, HttpServletRequest httpRequest) throws IOException
    {
    if (mTranslateHTTPMethods)
      {
      String httpMethod = httpRequest.getMethod();
      if ("GET".equals(httpMethod) || "HEAD".equals(httpMethod))
        {
        request.setVerb(INKFRequestReadOnly.VERB_SOURCE);
        }
      else if ("DELETE".equals(httpMethod))
        {
        request.setVerb(INKFRequestReadOnly.VERB_DELETE);
        }
      else if ("POST".equals(httpMethod) || "PUT".equals(httpMethod))
        {
        request.setVerb(INKFRequestReadOnly.VERB_SINK);
        Object content = this.getBodyContent(httpRequest);
        //TODO: Take out the if test when Tony fixes the Visualizer bug?
        if (content != null)
          {
          request.addPrimaryArgument(content);
          }
        else
          {
          // If we don't add a blank primary argument then the Visualizer will blow up on this request display
          request.addPrimaryArgument("");
          }
        }
      else
        {
        request.setVerb(INKFRequestReadOnly.VERB_SOURCE);
        context.logRaw(INKFRequestContext.LEVEL_DEBUG, "Unsupported HTTP method [" + httpMethod + "]");
        }
      }

    }


  private MimeTypePriority[] parseAcceptHeader(String acceptHeader)
    {
    MimeTypePriority[] mimeTypeOrderedList;

    String qValue;

    String[] typeSpecification = acceptHeader.split(",");
    mimeTypeOrderedList = new MimeTypePriority[typeSpecification.length];

    for (int i = 0; i < typeSpecification.length; i++)
      {
      String mimeType = typeSpecification[i].split(";")[0];
      try
        {
        String q = typeSpecification[i].split(";")[1];
        qValue = q.split("=")[1];
        }
      catch (java.lang.ArrayIndexOutOfBoundsException e)
        {
        qValue = "1.0";
        }
      mimeTypeOrderedList[i] = new MimeTypePriority(mimeType.trim(), qValue);
      }

    // Now sort the arrays with a simple Bubble Sort
    int n = mimeTypeOrderedList.length;
    for (int pass = 1; pass < n; pass++)
      {
      // Count how many times
      // This next loop becomes shorter and shorter
      for (int i = 0; i < n - pass; i++)
        {
        if (mimeTypeOrderedList[i].getPriority() < mimeTypeOrderedList[i + 1].getPriority())
          {
          // Exchange elements
          MimeTypePriority temp = mimeTypeOrderedList[i];
          mimeTypeOrderedList[i] = mimeTypeOrderedList[i + 1];
          mimeTypeOrderedList[i + 1] = temp;
          }
        }
      }
    return mimeTypeOrderedList;
    }
  }

class MimeTypePriority
  {
  private String mimeType;
  private String q;
  private float priority;
  private boolean endValue;

  public MimeTypePriority(String mimeType, String q)
    {
    this.mimeType = mimeType;
    this.q = q;
    if ("*".equals(mimeType) || "*/*".equals(mimeType))
      {
      endValue = true;
      }
    else
      {
      endValue = false;
      }

    if (q == null)
      {
      q = "1.0";
      priority = 1;
      }
    else
      {
      try
        {
        priority = Float.valueOf(q);
        }
      catch (Exception e)
        {
        q = "0.0";
        priority = 0;
        }
      }
    }


  public boolean isEndValue()
    {
    return endValue;
    }

  public float getPriority()
    {
    return priority;
    }

  public String getMimeType()
    {
    return mimeType;
    }

  public String getQ()
    {
    return q;
    }

  public String toString()
    {
    return "EndValue [" + isEndValue() + "], Priority [" + getPriority() + "], q [" + getQ() + "], MIME type [" + getMimeType() + "]";
    }
  }
